/*
 * Copyright (c) 2013 by Piero Orsoni (www.opbyte.it)
 * Based on the "sash" project by David I. Bell
 *
 * This file is part of cbmsh, a unix-like shell for commodore computers.
 *
 * Cbmsh is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cbmsh is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with cbmsh.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "cbmsh.h"
#include <cbm.h>
#include <conio.h>

#define	LISTSIZE	256	//8192

// Flags for the LS command.
#define	LSF_LONG	0x01
//#define	LSF_DIR		0x02
#define	LSF_USORT	0x08


// Data holding list of files.
static	char **list = NULL;
static	int	listSize = 0;
static	int	listUsed = 0;


// Local procedures.
static void	listFile(const char *name, char width);
static BOOL addListName(const char *fileName);
static void listAllFiles(char flags, char displayWidth, char displayHeight, char *line);
static void clearListNames(void);


void do_ls(int argc, const char **argv) {
	char flags = 0, line = 1, i, name[81], buf[20];
	char displayWidth, displayHeight;
	struct cbm_dirent dp;
	const char *cp;

	// clearListNames();
	argc--;
	argv++;

	while ((argc > 0) && (**argv == '-')) {
		cp = *argv++ + 1;
		argc--;

		while (*cp) switch (*cp++) {
			case 'l':	flags |= LSF_LONG; break;
			//case 'd':	flags |= LSF_DIR; break;
			case 'u':	flags |= LSF_USORT; break;
			default: printf("No opt -%c\n", cp[-1]); return;
		}
	}

	screensize(&displayWidth, &displayHeight);

	if (cbm_opendir(12, cbm_curunit()) != 0) {
		perror(".");
		return;
	}

	while (!cbm_readdir(12, &dp)) {
		if (dp.type == CBM_T_HEADER || !strcmp(dp.name, ".") || !strcmp(dp.name, "..")) continue;

		strcpy(name, dp.name);
		if (dp.type == CBM_T_DIR) strcat(name, "/");
		if (dp.type == CBM_T_PRG) strcat(name, "*");

		if (flags & LSF_LONG) {
			if (ltoa((long) dp.size * 256, buf, 10)) {
				// one less character, or it will \n and cause problems
				for (i = strlen(name) + strlen(buf) + 1; i < displayWidth; i++) strcat(name, " ");
				strcat(name, buf);
			}
		}

		if (argc == 0) {
			if (!addListName(name)) break;
		} else {
			for (i = 0; i < argc; i++) {
				if (!strcmp(dp.name, argv[i])) {		//  && (flags & LSF_DIR || dp.type != CBM_T_DIR)
					if (!addListName(name)) break;
				}
			}
		}
	}
	cbm_closedir(12);

	// List the files we collected in this directory, and then clear the list.
	listAllFiles(flags, displayWidth, displayHeight, &line);
	clearListNames();
}


/*
 * List all of the files in the current list of files.
 * The files are displayed according to the specified flags,
 * in the specified display width.
 */
static void listAllFiles(char flags, char displayWidth, char displayHeight, char *line) {
	char	fileWidth = 0, column = 0;
	char	len, ch;
	int		i;

	// Sort the files in the list.
	if (!(flags & LSF_USORT)) {
		qsort((void *) list, listUsed, sizeof(char *), nameSort);
	}

	// If we are showing the files in columns then calculate the maximum width of all of the file names
	if (!(flags & LSF_LONG)) {
		for (i = 0; i < listUsed; i++) {
			len = strlen(list[i]);
			if (fileWidth < len) fileWidth = len;
		}
		fileWidth++;
	}

	// Now list the fileNames.
	for (i = 0; i < listUsed; i++) {
		/*
		fp = open(name, O_RDONLY);
		if (fp <= 0) {
			perror(name);
			column = 0;
			(*line)++;
			continue;
		} else {
			close(fp);
		}
		*/

		/*
		cp = strrchr(name, '/');
		if (cp) cp++;
		else cp = name;
		*/

		// List the file in the next column or at the end of a line depending on the width left.
		if ((flags & LSF_LONG) || (column + fileWidth * 2 >= displayWidth)) {
			listFile(list[i], 0);
			column = 0;
			(*line)++;
		} else {
			listFile(list[i], fileWidth);
			column += fileWidth;
		}

		if (*line == displayHeight) {
			printf(STR_MORE);
			ch = cgetc();
			putchar('\n');
			if (ch == 'q' || ch == 'Q') break;
			*line = 1;
		}
	}

	// Terminate the last file name if necessary.
	if (column > 0) putchar('\n');
}


/*
 * Do a listing of a particular file name according to the flags.
 * The output is shown within the specified width if it is nonzero,
 * or on its own line if the width is zero.
 */
static void listFile(const char *name, char width) {
	char usedWidth = strlen(name);
	printf(name);

	// If no width was given then just end the line with a newline.
	if (width == 0) {
		//if (usedWidth < displayWidth)	used to check if last char, so no need to \n (causes havoc)
		putchar('\n');
	} else {
		// There is a width given. Print as many spaces as it takes to reach that width.
		while (usedWidth++ < width) putchar(' ');
	}
}


/*
 * Save a file name to the end of the static list, reallocating if necessary.
 * The file name is copied into allocated memory owned by the list.
 * Returns TRUE on success.
 */
static BOOL addListName(const char *fileName) {
	char **newList;

	// Reallocate the list if necessary.
	if (listUsed >= listSize) {
		newList = realloc(list, ((sizeof(char **)) *(listSize + LISTSIZE)));
		if (newList == NULL) {
			no_ram("name buf");
			return FALSE;
		}

		list = newList;
		listSize += LISTSIZE;
	}

	// Copy the file name into the next entry.
	list[listUsed] = strdup(fileName);

	if (list[listUsed] == NULL) {
		no_ram("file name");
		return FALSE;
	}

	// Increment the amount of space used.
	listUsed++;
	return TRUE;
}


// Free all of the names from the list of file names.
static void clearListNames(void) {
	while (listUsed > 0) {
		listUsed--;
		free(list[listUsed]);
	}
	listSize = 0;
	free(list);
	list = NULL;
}
