/*
 * Copyright (c) 2013 by Piero Orsoni (www.opbyte.it)
 * Based on the "sash" project by David I. Bell
 *
 * This file is part of cbmsh, a unix-like shell for commodore computers.
 *
 * Cbmsh is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cbmsh is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with cbmsh.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "cbmsh.h"
#include <cbm.h>
#include <device.h>
#include <peekpoke.h>


/*
 * A chunk of data.
 * Chunks contain data which is allocated as needed, but which is
 * not freed until all of the data needs freeing, such as at
 * the beginning of the next command.
 */
typedef	struct chunk CHUNK;
#define	CHUNK_INIT_SIZE	6		// 4

struct chunk {
	CHUNK *next;
	char data[CHUNK_INIT_SIZE];	// actually of varying length
};

static CHUNK *chunkList = NULL;


// Return TRUE if a fileName is a directory. Nonexistant files return FALSE.
/*
BOOL isDirectory(const char *name) {
	// TODO: GET TO KNOW IF IT IS A DIRECTORY!
	return (!strcmp(name, "."));
}
*/


/*
 * Copy one file to another, while possibly preserving its modes, times,
 * and modes.  Returns TRUE if successful, or FALSE on a failure with an
 * error message output.  (Failure is not indicted if the attributes cannot
 * be set.)
 */
BOOL copyFile(const char *srcName, const char *destName) {
	char buf[BUF_SIZE];
	int rfd, wfd, rcc;

	rfd = open(srcName, O_RDONLY);
	if (rfd < 0) {
		perror(srcName);
		return FALSE;
	}

	wfd = creat(destName, O_CREAT);
	if (wfd < 0) {
		perror(destName);
		close(rfd);
		return FALSE;
	}

	while ((rcc = read(rfd, buf, sizeof(buf))) > 0) {
		if (fullWrite(wfd, buf, rcc) < 0) goto error_exit;
	}

	if (rcc < 0) {
		perror(srcName);
		goto error_exit;
	}

	close(rfd);
	if (close(wfd) < 0) {
		perror(destName);
		return FALSE;
	}
	return TRUE;

error_exit:
	close(rfd);
	close(wfd);
	return FALSE;
}


/*
 * Build a path name from the specified directory name and file name.
 * If the directory name is NULL, then the original fileName is returned.
 * The built path is in a static area, and is overwritten for each call.
 */
/*
const char *buildName(const char *dirName, const char *fileName) {
	const char *cp;
	static char buf[PATH_LEN * 2 + 1];

	if ((dirName == NULL) || (*dirName == '\0')) return fileName;
	cp = strrchr(fileName, '/');
	if (cp) fileName = cp + 1;

	strcpy(buf, dirName);
	strcat(buf, "/");
	strcat(buf, fileName);
	return buf;
}
*/


/*
 * Expand the wildcards in a fileName wildcard pattern, if any.
 * Returns an argument list with matching fileNames in sorted order.
 * The expanded names are stored in memory chunks which can later all
 * be freed at once.  The returned list is only valid until the next
 * call or until the next command.  Returns zero if the name is not a
 * wildcard, or returns the count of matched files if the name is a
 * wildcard and there was at least one match, or returns -1 if either
 * no fileNames matched or there was an allocation error.
 */
int expandWildCards(const char *fileNamePattern, const char ***retFileTable) {
	const char *cp1, *cp2, *cp3;
	char *str;
	struct cbm_dirent dp;
	char **newFileTable;
	int	newFileTableSize;

	static int fileCount;
	static int fileTableSize;
	static char **fileTable = NULL;

	// Clear the return values until we know their final values.
	fileCount = 0;
	*retFileTable = NULL;

	// Scan the file name pattern for any wildcard characters.
	cp1 = strchr(fileNamePattern, '*');
	cp2 = strchr(fileNamePattern, '?');
	cp3 = strchr(fileNamePattern, '[');

	// If there are no wildcard characters then return zero to indicate that there was actually no wildcard pattern.
	if ((cp1 == NULL) && (cp2 == NULL) && (cp3 == NULL)) return 0;

	// Open the directory containing the files to be checked.
	if (cbm_opendir(13, cbm_curunit()) != 0)	{
		perror(".");
		return -1;
	}

	// Find all of the files in the directory and check them against the wildcard pattern.
	while (!cbm_readdir(13, &dp)) {
		// Skip the current and parent directories.
		if (dp.type == CBM_T_HEADER || !strcmp(dp.name, ".") || !strcmp(dp.name, "..")) continue;

		// If the file name doesn't match the pattern then skip it.
		if (!match(dp.name, fileNamePattern)) continue;

		// This file name is selected. See if we need to reallocate the file name table.
		if (fileCount >= fileTableSize) {
			// Increment the file table size and reallocate it.
			newFileTableSize = fileTableSize + EXPAND_ALLOC;
			newFileTable = (char **) realloc((char *) fileTable, (newFileTableSize * sizeof(char *)));

			if (newFileTable == NULL) {
				no_ram("file list");
				cbm_closedir(13);
				return -1;
			}

			fileTable = newFileTable;
			fileTableSize = newFileTableSize;
		}

		// Allocate space for storing the file name in a chunk.
		str = getChunk(strlen(dp.name) + 1);

		if (str == NULL) {
			no_ram("file name");
			cbm_closedir(13);
			return -1;
		}

		// Save the file name in the chunk.
		strcpy(str, dp.name);

		// Save the allocated file name into the file table.
		fileTable[fileCount++] = str;
	}

	// Close the directory and check for any matches.
	cbm_closedir(13);

	if (fileCount == 0) {
		// printf("No matches\n");
		return -1;
	}

	// Sort the list of file names.
	qsort((void *) fileTable, fileCount, sizeof(char *), nameSort);

	// Return the file list and count.
	*retFileTable = (const char **) fileTable;
	return fileCount;
}


// Sort routine for list of fileNames.
int nameSort(const void *p1, const void *p2) {
	return strcmp(*((const char**) p1), *((const char**) p2));
}


/*
 * Routine to see if a text string is matched by a wildcard pattern.
 * Returns TRUE if the text is matched, or FALSE if it is not matched
 * or if the pattern is invalid.
 *  *		matches zero or more characters
 *  ?		matches a single character
 *  [abc]	matches 'a', 'b' or 'c'
 *  \c		quotes character c
 *  Adapted from code written by Ingo Wilken.
 */
BOOL match(const char *text, const char *pattern) {
	const char *retryPat = NULL;
	const char *retryText = NULL;
	BOOL found;
	char ch;

	while (*text || *pattern)	{
		ch = *pattern++;

		switch (ch) {
			case '*':
				retryPat = pattern;
				retryText = text;
				break;
			case '[':
				found = FALSE;

				while ((ch = *pattern++) != ']') {
					if (ch == '\\') ch = *pattern++;
					if (ch == '\0') return FALSE;
					if (*text == ch) found = TRUE;
				}

				if (!found) {
					pattern = retryPat;
					text = ++retryText;
				}
				/* fall into next case */
			case '?':
				if (*text++ == '\0') return FALSE;
				break;
			case '\\':
				ch = *pattern++;
				if (ch == '\0') return FALSE;
				/* fall into next case */
			default:
				if (*text == ch) {
					if (*text) text++;
					break;
				}
				if (*text) {
					pattern = retryPat;
					text = ++retryText;
					break;
				}
				return FALSE;
		}

		if (pattern == NULL) return FALSE;
	}
	return TRUE;
}


/*
 * Take a command string and break it up into an argc, argv list while
 * handling quoting and wildcards.  The returned argument list and
 * strings are in static memory, and so are overwritten on each call.
 * The argument list is ended with a NULL pointer for convenience.
 * Returns TRUE if successful, or FALSE on an error with a message
 * already output.
 */
BOOL makeArgs(const char *cmd, int *retArgc, const char ***retArgv) {
	const char *argument;
	char *cp, *cpOut, *newStrings;
	const char **newArgTable;
	int newArgTableSize;
	int fileCount;
	char ch, quote;
	BOOL quotedWildCards, unquotedWildCards;
	const char **fileTable = NULL;
	int len = strlen(cmd) + 1;

	static int stringsLength;
	static char *strings = NULL;
	static int argCount;
	static int argTableSize;
	static const char **argTable = NULL;

	// Clear the returned values until we know them.
	argCount = 0;
	*retArgc = 0;
	*retArgv = NULL;

	// Copy the command string into a buffer that we can modify, reallocating it if necessary.
	if (len > stringsLength) {
		newStrings = realloc(strings, len);
		if (newStrings == NULL) {
			no_ram("string");
			return FALSE;
		}
		strings = newStrings;
		stringsLength = len;
	}

	memcpy(strings, cmd, len);
	cp = strings;

	// Keep parsing the command string as long as there are any arguments left.
	while (*cp)	{
		// Save the beginning of this argument.
		argument = cp;
		cpOut = cp;

		// Reset quoting and wildcarding for this argument.
		quote = '\0';
		quotedWildCards = FALSE;
		unquotedWildCards = FALSE;

		/*
		 * Loop over the string collecting the next argument while
		 * looking for quoted strings or quoted characters, and
		 * remembering whether there are any wildcard characters
		 * in the argument.
		 */
		while (*cp) {
			ch = *cp++;

			// If we are not in a quote and we see a blank then this argument is done.
			if (isBlank(ch) && (quote == '\0')) break;

			// If we see a backslash then accept the next character no matter what it is.
			if (ch == '\\') {
				ch = *cp++;

				// Make sure there is a next character.
				if (ch == '\0') {
					printf("Bad quote\n");
					return FALSE;
				}

				// Remember whether the quoted character is a wildcard.
				if (isWildCard(ch)) quotedWildCards = TRUE;
				*cpOut++ = ch;
				continue;
			}

			/*
			 * If we see one of the wildcard characters then
			 * remember whether it was seen inside or outside
			 * of quotes.
			 */
			if (isWildCard(ch))	{
				if (quote) quotedWildCards = TRUE;
				else unquotedWildCards = TRUE;
			}

			/*
			 * If we were in a quote and we saw the same quote
			 * character again then the quote is done.
			 */
			if (ch == quote) {
				quote = '\0';
				continue;
			}

			/*
			 * If we weren't in a quote and we see either type
			 * of quote character, then remember that we are
			 * now inside of a quote.
			 */
			if ((quote == '\0') && ((ch == '\'') || (ch == '"'))) {
				quote = ch;
				continue;
			}

			// Store the character.
			*cpOut++ = ch;
		}

		// Make sure that quoting is terminated properly.
		if (quote) {
			printf("Bad quote\n");		// Unmatched quote char
			return FALSE;
		}

		/*
		 * Null terminate the argument if it had shrunk, and then
		 * skip over all blanks to the next argument, nulling them
		 * out too.
		 */
		if (cp != cpOut) *cpOut = '\0';
		while (isBlank(*cp)) *cp++ = '\0';

		/*
		 * If both quoted and unquoted wildcards were used then
		 * complain since we don't handle them properly.
		 */
		/*
		if (quotedWildCards && unquotedWildCards) {
			printf("Mixed quotes\n");
			return FALSE;
		}
		*/

		/*
		 * Expand the argument into the matching filenames or accept
		 * it as is depending on whether there were any unquoted
		 * wildcard characters in it.
		 */
		if (unquotedWildCards) {
			// Expand the argument into the matching filenames.
			fileCount = expandWildCards(argument, &fileTable);

			// Return an error if the wildcards failed to match.
			if (fileCount < 0) return FALSE;

			if (fileCount == 0) {
				printf("Wildcard err\n");		// Wildcard expansion error
				return FALSE;
			}
		} else {
			// Set up to only store the argument itself.
			fileTable = &argument;
			fileCount = 1;
		}

		// Now reallocate the argument table to hold the file name.
		if (argCount + fileCount >= argTableSize) {
			newArgTableSize = argCount + fileCount + 1;
			newArgTable = (const char **) realloc(argTable, (sizeof(const char *) *newArgTableSize));

			if (newArgTable == NULL) {
				no_ram("arg list");
				return FALSE;
			}

			argTable = newArgTable;
			argTableSize = newArgTableSize;
		}

		// Copy the new arguments to the end of the old ones.
		memcpy((void *) &argTable[argCount], (const void *) fileTable, (sizeof(const char **) *fileCount));

		// Add to the argument count.
		argCount += fileCount;
	}

	// Null terminate the argument list and return it.
	argTable[argCount] = NULL;
	*retArgc = argCount;
	*retArgv = argTable;
 	return TRUE;
}


/*
 * Make a NULL-terminated string out of an argc, argv pair.
 * Returns TRUE if successful, or FALSE if the string is too long,
 * with an error message given.  This does not handle spaces within
 * arguments correctly.
 */
BOOL makeString(int argc, const char **argv, char *buf, int bufLen) {
	int	len;

	while (argc-- > 0) {
		len = strlen(*argv);
		if (len >= bufLen) {
			printf("Args too long\n");
			return FALSE;
		}

		strcpy(buf, *argv++);
		buf += len;
		bufLen -= len;

		if (argc) *buf++ = ' ';
		bufLen--;
	}

	*buf = '\0';
	return TRUE;
}


/*
 * Allocate a chunk of memory (like malloc).
 * The difference, though, is that the memory allocated is put on a
 * list of chunks which can be freed all at one time.  You CAN NOT free
 * an individual chunk.
 */
char *getChunk(int size) {
	CHUNK *chunk;

	if (size < CHUNK_INIT_SIZE) size = CHUNK_INIT_SIZE;
	chunk = (CHUNK *) malloc(size + sizeof(CHUNK) - CHUNK_INIT_SIZE);
	if (chunk == NULL) return NULL;

	chunk->next = chunkList;
	chunkList = chunk;
	return chunk->data;
}


/*
 * Duplicate a string value using the chunk allocator.
 * The returned string cannot be individually freed, but can only be freed
 * with other strings when freeChunks is called.  Returns NULL on failure.
 */
char *chunkstrdup(const char * str) {
	int	len = strlen(str) + 1;
	char *newStr = getChunk(len);
	if (newStr) memcpy(newStr, str, len);
	return newStr;
}


/*
 * Free all chunks of memory that had been allocated since the last
 * call to this routine.
 */
void freeChunks(void) {
	CHUNK *chunk;
	while (chunkList)	{
		chunk = chunkList;
		chunkList = chunk->next;
		free((char *) chunk);
	}
	chunkList = NULL;
}


/*
 * Write all of the supplied buffer out to a file. This does multiple writes as necessary.
 * Returns the amount written, or -1 on an error.
 */
int fullWrite(int fd, const char *buf, int len) {
	int	cc, total = 0;

	while (len > 0) {
		cc = write(fd, buf, len);
		if (cc < 0) return -1;

		buf += cc;
		total += cc;
		len -= cc;
	}
	return total;
}


/*
 * Read all of the supplied buffer from a file. This does multiple reads as necessary.
 * Returns the amount read, or -1 on an error. A short read is returned on an end of file.
 */
int fullRead(int fd, char *buf, int len) {
	int	cc, total = 0;

	while (len > 0) {
		cc = read(fd, buf, len);
		if (cc < 0) return -1;
		if (cc == 0) break;

		buf += cc;
		total += cc;
		len -= cc;
	}
	return total;
}


char cbm_curunit() {
	char buf[3];
	if (getcwd(buf, 3) != NULL) return atoi(buf);
	return INVALID_DEVICE;
	/*
	// should work but probably clashes with cc65 library routines
	char u = PEEK(186);
	if (u < 8) u = 8;
	return u;
	*/
}


char dos_command(char *cmd, char **mes) {
	static char message[40];
	char ret = 0;

  if (cbm_open(15, cbm_curunit(), 15, cmd) != 0) {
		printf("Can't run DOS\n");
		ret = -1;		// _oserror
  } else {
    if (cbm_read(15, message, 39) > 0) {
			//message[size] = '\0';
			ret = atoi(strtok(message, ","));
    }

		if (ret > 0) *mes = strtok(NULL, ",");
		cbm_close(15);
  }
  return ret;
}


char cbm_command(char *cmd) {
	char *mes, ret;
	if ((ret = dos_command(cmd, &mes)) > 0) {
		printf(mes);
		putchar('\n');
  }
  return ret;
}


char cbm_drivetype() {
	char *mes, ret = 0;
	if (dos_command("ui", &mes) > 0) {
		if (strstr(mes, "vice")) ret = 1;
  }
  return ret;
}
